@@ -4,8 +4,8 @@ import tinychain.error
 import tinychain.graph
 import tinychain.graph.edge
 import tinychain.host
-import tinychain.ml.cnn
-import tinychain.ml.dnn
+import tinychain.ml.nn
+import tinychain.ml.optimizer
 import tinychain.op
 import tinychain.ref
 
