@@ -1,9 +1,10 @@
 """An n-dimensional array of numbers."""
 
 from tinychain import ref
-from tinychain.state import Map, State, Stream, Tuple
+from tinychain.reflect import is_ref
+from tinychain.state import Class, State, Stream, Tuple
 from tinychain.util import form_of, to_json, uri, URI
-from tinychain.value import Bool, I16, F32, Number, UInt
+from tinychain.value import Bool, I16, F32, F64, Number, UInt
 
 from .bound import Range
 from .collection import Collection
@@ -20,6 +21,14 @@ class Schema(object):
         self.shape = shape
         self.dtype = dtype
 
+    def __getitem__(self, i):
+        if i == 0:
+            return self.shape
+        elif i == 1:
+            return self.dtype
+        else:
+            raise KeyError(f"Tensor schema (shape, dtype) has no element {i}")
+
     def __json__(self):
         return to_json([self.shape, self.dtype])
 
@@ -36,7 +45,8 @@ class Tensor(Collection):
     def create(cls, shape, dtype=F32):
         """Create a new, empty `Tensor`. Call this method to initialize a persistent `Tensor` in a `Chain`."""
 
-        return cls(ref.Get(cls, Schema(shape, dtype)))
+        schema = Schema(shape, dtype)
+        return cls(Create(cls, schema, schema))
 
     @classmethod
     def load(cls, shape, dtype, data):
@@ -71,6 +81,9 @@ class Tensor(Collection):
     def __add__(self, other):
         return self.add(other)
 
+    def __abs__(self):
+        return self.abs()
+
     def __radd__(self, other):
         return self.add(other)
 
@@ -116,6 +129,16 @@ class Tensor(Collection):
     def __rtruediv__(self, other):
         return other * (self.pow(-1))
 
+    def __ref__(self, name):
+        form = form_of(self)
+        if hasattr(form, "schema"):
+            ref_form = TypeRef(URI(name), form.schema)
+            return self.__class__(ref_form)
+        elif hasattr(form, "__ref__"):
+            return self.__class__(get_ref(form, name))
+        else:
+            return self.__class__(URI(name))
+
     def abs(self):
         """Return the element-wise absolute value of this `Tensor`."""
 
@@ -134,7 +157,7 @@ class Tensor(Collection):
     def add(self, other):
         """Return the element-wise sum of this `Tensor` and another `Tensor` or `Number`."""
 
-        return self._post("add", Map(r=other), Tensor)
+        return self._post("add", {"r": other}, Tensor)
 
     def all(self):
         """Return `True` if all elements in this `Tensor` are nonzero."""
@@ -146,6 +169,14 @@ class Tensor(Collection):
 
         return self._get("any", rtype=Bool)
 
+    def argmax(self, axis=None):
+        """Return the indices of the maximum values along the given `axis` of this `Tensor`.
+
+        If no `axis` is given, the total offset will be returned.
+        """
+
+        return self._get("argmax", axis, self.__class__)
+
     def asin(self):
         """Return the element-wise arcsine of this `Tensor`."""
 
@@ -189,7 +220,13 @@ class Tensor(Collection):
     def div(self, other):
         """Divide this `Tensor` by another `Tensor` or `Number`, broadcasting if necessary."""
 
-        return self._post("div", Map(r=other), Tensor)
+        return self._post("div", {"r": other}, Tensor)
+
+    @property
+    def dtype(self):
+        """Return the data type of this `Tensor`."""
+
+        return self._get("dtype", rtype=Class)
 
     def flip(self, axis):
         """Flip the elements in this `Tensor` along the specified `axis`."""
@@ -199,7 +236,7 @@ class Tensor(Collection):
     def eq(self, other):
         """Return a boolean `Tensor` with element-wise equality values."""
 
-        return self._post("eq", Map(r=other), Tensor)
+        return self._post("eq", {"r": other}, Tensor)
 
     def exp(self):
         """Raise `e` to the power of this `Tensor`."""
@@ -214,27 +251,36 @@ class Tensor(Collection):
     def gt(self, other):
         """Return a boolean `Tensor` with element-wise greater-than values."""
 
-        return self._post("gt", Map(r=other), self.__class__)
+        return self._post("gt", {"r": other}, self.__class__)
 
     def gte(self, other):
         """Return a boolean `Tensor` with element-wise greater-or-equal values."""
 
-        return self._post("gte", Map(r=other), Tensor)
+        return self._post("gte", {"r": other}, Tensor)
+
+    def log(self, base=None):
+        """
+        Return the logarithm of this `Tensor`.
+
+        If no `base` is specified, this will return the natural logarithm (base e).
+        """
+
+        return self._get("log", base, self.__class__)
 
     def lt(self, other):
         """Return a boolean `Tensor` with element-wise less-than values."""
 
-        return self._post("lt", Map(r=other), self.__class__)
+        return self._post("lt", {"r": other}, self.__class__)
 
     def lte(self, other):
         """Return a boolean `Tensor` with element-wise less-or-equal values."""
 
-        return self._post("lte", Map(r=other), Tensor)
+        return self._post("lte", {"r": other}, Tensor)
 
     def logical_and(self, other):
         """Return a boolean `Tensor` with element-wise logical and values."""
 
-        return self._post("and", Map(r=other), self.__class__)
+        return self._post("and", {"r": other}, self.__class__)
 
     def logical_not(self):
         """Return a boolean `Tensor` with element-wise logical not values."""
@@ -244,12 +290,12 @@ class Tensor(Collection):
     def logical_or(self, other):
         """Return a boolean `Tensor` with element-wise logical or values."""
 
-        return self._post("or", Map(r=other), Tensor)
+        return self._post("or", {"r": other}, Tensor)
 
     def logical_xor(self, other):
         """Return a boolean `Tensor` with element-wise logical xor values."""
 
-        return self._post("xor", Map(r=other), Tensor)
+        return self._post("xor", {"r": other}, Tensor)
 
     @property
     def ndim(self):
@@ -257,20 +303,31 @@ class Tensor(Collection):
 
         return self._get("ndim", rtype=UInt)
 
+    def mean(self, axis=None):
+        """
+        Return the average of this `Tensor` along the given `axis`,
+        or the average of the entire `Tensor` if no axis is given.
+        """
+
+        if axis is None:
+            return self.sum() / self.size
+        else:
+            return self.sum(axis) / self.shape[axis]
+
     def mul(self, other):
         """Multiply this `Tensor` by another `Tensor` or `Number`, broadcasting if necessary."""
 
-        return self._post("mul", Map(r=other), self.__class__)
+        return self._post("mul", {"r": other}, self.__class__)
 
     def ne(self, other):
         """Return a boolean `Tensor` with element-wise not-equal values."""
 
-        return self._post("ne", Map(r=other), self.__class__)
+        return self._post("ne", {"r": other}, self.__class__)
 
     def pow(self, other):
         """Raise this `Tensor` to the given power."""
 
-        return self._post("pow", Map(r=other), self.__class__)
+        return self._post("pow", {"r": other}, self.__class__)
 
     def product(self, axis=None):
         """Calculate the product of this `Tensor` along the given `axis`, or the total product if no axis is given."""
@@ -283,12 +340,22 @@ class Tensor(Collection):
 
         return self._get("reshape", shape, self.__class__)
 
-    # TODO: use a custom shape object to support `type(x.shape[0]) == UInt`
+    def round(self):
+        """Round this `Tensor` to the nearest integer, element-wise."""
+
+        return self._get("round", rtype=self.__class__)
+
     @property
     def shape(self):
         """Return the shape of this `Tensor`."""
 
-        return self._get("shape", rtype=Tuple)
+        form = form_of(self)
+        if hasattr(form, "schema"):
+            shape = form.schema[0]
+            shape = shape if isinstance(shape, Tuple) else Tuple(shape)
+            return shape
+        else:
+            return self._get("shape", rtype=Tuple)
 
     def sin(self):
         """Return the element-wise sine of this `Tensor`."""
@@ -300,6 +367,12 @@ class Tensor(Collection):
 
         return self._get("sinh", rtype=self.__class__)
 
+    @property
+    def size(self):
+        """Return the size of this `Tensor` (the product of its `shape`)."""
+
+        return self._get("size", rtype=UInt)
+
     def split(self, num_or_size_splits, axis=0):
         """
         Split this `Tensor` into multiple slices along the given `axis`.
@@ -312,12 +385,12 @@ class Tensor(Collection):
         `self.shape[axis]` then a `BadRequest` error will be raised.
         """
 
-        return self._get("split", (num_or_size_splits, axis), self.__class__)
+        return self._get("split", (num_or_size_splits, axis), Tuple)
 
     def sub(self, other):
         """Subtract another `Tensor` or `Number` from this one, broadcasting if necessary."""
 
-        return self._post("sub", Map(r=other), Tensor)
+        return self._post("sub", {"r": other}, Tensor)
 
     def sum(self, axis=None):
         """Calculate the sum of this `Tensor` along the given `axis`, or the total sum if no axis is given."""
@@ -368,7 +441,9 @@ class Dense(Tensor):
         evenly distributed between `start` and `stop`.
         """
 
-        return cls(ref.Get(uri(cls) + "/range", (shape, start, stop)))
+        dtype = type(start) if isinstance(start, Number) else Number
+        schema = Schema(shape, dtype)
+        return cls(Create(uri(cls) + "/range", (shape, start, stop), schema))
 
     @classmethod
     def concatenate(cls, tensors, axis=None):
@@ -384,7 +459,46 @@ class Dense(Tensor):
     def constant(cls, shape, value):
         """Return a `Dense` tensor filled with the given `value`."""
 
-        return cls(ref.Get(uri(cls) + "/constant", (shape, value)))
+        dtype = type(value) if isinstance(value, Number) else Number
+        schema = Schema(shape, dtype)
+        return cls(Create(uri(cls) + "/constant", (shape, value), schema))
+
+    @classmethod
+    def load(cls, shape, dtype, data):
+        """
+        Load a `Dense` tensor from an existing data set.
+
+        Example: `tc.tensor.Dense.load([2, 2], tc.i32, [1, 2, 3, 4])`
+        """
+
+        if is_ref(shape) or is_ref(dtype):
+            raise ValueError(f"cannot load schema ({shape}, {dtype}) (consider calling `copy_from` instead)")
+
+        if is_ref(data):
+            raise ValueError(f"cannot load data {data} (consider calling `copy_from` instead)")
+
+        schema = Schema(shape, dtype)
+
+        class Load(cls):
+            def __init__(self, put_op):
+                cls.__init__(self, put_op)
+
+            @property
+            def dtype(self):
+                return schema[1]
+
+            @property
+            def schema(self):
+                return schema
+
+            @property
+            def shape(self):
+                return schema[0]
+
+            def __ref__(self, name):
+                return cls(URI(name))
+
+        return Load(ref.Put(cls, schema, data))
 
     @classmethod
     def ones(cls, shape, dtype=F32):
@@ -394,7 +508,7 @@ class Dense(Tensor):
         If `dtype` is not specified, the data type will be :class:`F32`.
         """
 
-        return cls.constant(shape, dtype(1))
+        return cls.constant(shape, Number(1).cast(dtype))
 
     @classmethod
     def zeros(cls, shape, dtype=F32):
@@ -404,7 +518,27 @@ class Dense(Tensor):
         If `dtype` is not specified, the data type will be :class:`F32`.
         """
 
-        return cls.constant(shape, dtype(0))
+        return cls.constant(shape, Number(0).cast(dtype))
+
+    @classmethod
+    def random_normal(cls, shape, mean=0.0, std=1.0):
+        """Return a `Dense` tensor filled with a random normal distribution of `F64` s."""
+
+        schema = Schema(shape, F64)
+        args = {"shape": shape, "mean": mean, "std": std}
+        return cls(Distribution(uri(cls) + "/random/normal", args, schema))
+
+    @classmethod
+    def random_uniform(cls, shape):
+        """Return a `Dense` tensor filled with a uniform random distribution of `F64` s."""
+
+        schema = Schema(shape, F64)
+        return cls(Create(uri(cls) + "/random/uniform", shape, schema))
+
+    def argsort(self):
+        """Return the coordinates needed to sort this `Tensor`."""
+
+        return self._get("argsort", rtype=self.__class__)
 
     def elements(self, bounds):
         """Return a :class:`Stream` of the :class:`Number` elements of this `Dense` tensor."""
@@ -497,3 +631,24 @@ def _handle_bounds(bounds):
     return [
         Range.from_slice(x) if isinstance(x, slice)
         else x for x in bounds]
+
+
+class Create(ref.Get):
+    def __init__(self, subject, args, schema):
+        ref.Get.__init__(self, subject, args)
+        self.schema = schema
+
+
+class Distribution(ref.Get):
+    def __init__(self, subject, args, schema):
+        ref.Post.__init__(self, subject, args)
+        self.schema = schema
+
+
+class TypeRef(ref.Ref):
+    def __init__(self, name, schema):
+        self.__uri__ = name
+        self.schema = schema
+
+    def __json__(self):
+        return to_json(self.__uri__)
