@@ -3,8 +3,10 @@ Reference types.
 :class:`After`, :class:`Case`, :class:`If`, and :class:`While` are available in the top-level namespace.
 """
 
-from tinychain.reflect import is_conditional
-from tinychain.util import deanonymize, form_of, get_ref, hex_id, requires, to_json, uri, URI
+import logging
+
+from tinychain.reflect import is_conditional, is_op, is_ref
+from tinychain.util import deanonymize, form_of, get_ref, hex_id, to_json, uri, URI
 
 
 class Ref(object):
@@ -32,11 +34,8 @@ class After(Ref):
         self.when = when
         self.then = then
 
-    def __deps__(self):
-        deps = set()
-        deps.update(requires(self.when))
-        deps.update(requires(self.then))
-        return deps
+    def __dbg__(self):
+        return [self.when, self.then]
 
     def __json__(self):
         return {str(uri(self)): to_json([self.when, self.then])}
@@ -74,12 +73,8 @@ class Case(Ref):
         self.switch = switch
         self.case = case
 
-    def __deps__(self):
-        deps = set()
-        deps.update(requires(self.cond))
-        deps.update(requires(self.switch))
-        deps.update(requires(self.case))
-        return deps
+    def __dbg__(self):
+        return [self.cond, self.switch, self.case]
 
     def __json__(self):
         return {str(uri(self)): to_json([self.cond, self.switch, self.case])}
@@ -112,12 +107,8 @@ class If(Ref):
         self.then = then
         self.or_else = or_else
 
-    def __deps__(self):
-        deps = set()
-        deps.update(requires(self.cond))
-        deps.update(requires(self.then))
-        deps.update(requires(self.or_else))
-        return deps
+    def __dbg__(self):
+        return [self.cond, self.then, self.or_else]
 
     def __json__(self):
         return {str(uri(self)): to_json([self.cond, self.then, self.or_else])}
@@ -136,14 +127,11 @@ class While(Ref):
     A flow control operator to execute a closure repeatedly until a condition is met.
 
     Args:
-        cond (Bool or Ref): The condition which determines which branch to execute.
+        cond (Post): The condition which determines whether to terminate the loop.
 
-        then (State or Ref): The State to resolve if `cond` resolves to `True`.
+        step (Post): One cycle of this `While` loop.
 
-        or_else (State or Ref): The State to resolve if `cond` resolves to `False`.
-
-    Raises:
-        `BadRequestError` if `cond` does not resolve to a :class:`Bool` or in case of a nested conditional
+        state (State): The initial state of the loop.
     """
 
     __uri__ = uri(Ref) + "/while"
@@ -153,12 +141,8 @@ class While(Ref):
         self.op = op
         self.state = state
 
-    def __deps__(self):
-        deps = set()
-        deps.update(requires(self.cond))
-        deps.update(requires(self.op))
-        deps.update(requires(self.state))
-        return deps
+    def __dbg__(self):
+        return [self.cond, self.op, self.state]
 
     def __json__(self):
         return {str(uri(self)): to_json([self.cond, self.op, self.state])}
@@ -185,24 +169,22 @@ class With(Ref):
         self.capture = []
 
         for ref in capture:
-            captured = uri(ref).subject()
-
-            if captured is None:
+            id = uri(ref).id()
+            if id is None:
                 raise ValueError(f"With can only capture states with an ID in the current context, not {ref}")
 
-            self.capture.append(captured)
+            self.capture.append(URI(id))
 
         self.op = op
 
-    def __deps__(self):
-        return set(self.capture)
+    def __dbg__(self):
+        return [self.capture, self.op]
 
     def __json__(self):
         return {str(uri(self)): to_json([self.capture, self.op])}
 
-    def __ns__(self, cxt):
-        deanonymize(self.capture, cxt)
-        deanonymize(self.op, cxt)
+    def __ns__(self, _cxt):
+        pass
 
     def __ref__(self, name):
         return get_ref(self.op, name)
@@ -217,11 +199,9 @@ class Op(Ref):
         self.subject = subject
         self.args = args
 
-    def __deps__(self):
-        deps = set()
-        deps.update(requires(self.subject))
-        deps.update(requires(self.args))
-        return deps
+    def __dbg__(self):
+        subject = [self.subject] if is_ref(self.subject) else []
+        return subject + list(self.args)
 
     def __json__(self):
         if hasattr(self.subject, "__form__"):
@@ -256,8 +236,8 @@ class Get(Op):
 
     def __json__(self):
         if isinstance(self.subject, Ref):
-            subject = self.subject
-            is_scalar = True
+            subject = uri(self.subject)
+            is_scalar = False
         else:
             subject = uri(self.subject)
             if subject is None:
@@ -265,7 +245,7 @@ class Get(Op):
 
             is_scalar = subject.startswith("/state/scalar")
 
-        if is_scalar:
+        if is_scalar and not is_ref(self.args):
             (value,) = self.args
             return {str(subject): to_json(value)}
         else:
@@ -274,6 +254,14 @@ class Get(Op):
     def __repr__(self):
         return f"GET Op ref {self.subject} {self.args}"
 
+    def __ns__(self, cxt):
+        super().__ns__(cxt)
+
+        if is_op_ref(self.args):
+            (key,) = self.args
+            log_anonymous(key)
+            self.args = (reference(cxt, key),)
+
 
 class Put(Op):
     """
@@ -295,6 +283,20 @@ class Put(Op):
     def __repr__(self):
         return f"PUT Op ref {self.subject} {self.args}"
 
+    def __ns__(self, cxt):
+        super().__ns__(cxt)
+
+        if is_op_ref(self.args):
+            key, value = self.args
+
+            log_anonymous(key)
+            key = reference(cxt, key)
+
+            log_anonymous(value)
+            value = reference(cxt, value)
+
+            self.args = (key, value)
+
 
 class Post(Op):
     """
@@ -309,11 +311,30 @@ class Post(Op):
     __uri__ = uri(Op) + "/post"
 
     def __init__(self, subject, args):
+        if not hasattr(args, "__iter__"):
+            raise ValueError("POST Op ref requires named parameters (try using a Python dict)")
+
         Op.__init__(self, subject, args)
 
     def __repr__(self):
         return f"POST Op ref {self.subject} {self.args}"
 
+    def __ns__(self, cxt):
+        super().__ns__(cxt)
+
+        if not isinstance(self.args, dict):
+            raise ValueError(f"POST arguments must be a Python dict, not {self.args}")
+
+        args = {}
+        for name, arg in self.args.items():
+            if is_op_ref(arg):
+                log_anonymous(arg)
+                args[name] = reference(cxt, arg)
+            else:
+                args[name] = arg
+
+        self.args = args
+
 
 class Delete(Op):
     """
@@ -336,9 +357,16 @@ class Delete(Op):
     def __repr__(self):
         return f"DELETE Op ref {self.subject} {self.args}"
 
+    def __ns__(self, cxt):
+        super().__ns__(cxt)
+
+        if is_op_ref(self.args):
+            log_anonymous(self.args)
+            self.args = reference(cxt, self.args)
+
 
 class MethodSubject(object):
-    def __init__(self, subject, method_name):
+    def __init__(self, subject, method_name=""):
         if uri(subject).startswith('$'):
             self.__uri__ = uri(subject).append(method_name)
         else:
@@ -347,24 +375,38 @@ class MethodSubject(object):
         self.subject = subject
         self.method_name = method_name
 
-    def __deps__(self):
-        return requires(self.subject)
+    def __dbg__(self):
+        subject = [self.subject] if is_ref(self.subject) else []
+        return subject + self.args
+
+    def __dbg__(self):
+        return [self.subject]
 
     def __ns__(self, cxt):
+        name = f"{self.subject.__class__.__name__}_{hex_id(self.subject)}"
+        auto_uri = URI(name).append(self.method_name)
+
         if uri(self):
-            return
-
-        deanonymize(self.subject, cxt)
-
-        if uri(self.subject).startswith("/state"):
-            name = f"{self.subject.__class__.__name__}_{hex_id(self.subject)}"
-            self.__uri__ = URI(name).append(self.method_name)
-
-            if not cxt.is_defined(name):
+            if uri(self) == auto_uri and name not in cxt:
+                logging.debug(f"auto-assigning name {name} to {self.subject} in {cxt}")
                 setattr(cxt, name, self.subject)
+            else:
+                return
+
+        elif uri(self.subject).startswith("/state"):
+            self.__uri__ = auto_uri
+
+            if name not in cxt:
+                logging.debug(f"auto-assigning name {name} to {self.subject} in {cxt}")
+                setattr(cxt, name, self.subject)
+
+            deanonymize(self.subject, cxt)
+
         else:
+            deanonymize(self.subject, cxt)
             self.__uri__ = uri(self.subject).append(self.method_name)
 
+
     def __json__(self):
         if self.__uri__ is None:
             raise ValueError(f"cannot call method {self.method_name} on an anonymous subject {self.subject}")
@@ -377,12 +419,47 @@ class MethodSubject(object):
         else:
             return str(uri(self))
 
-    def assign_uri(self, subject_uri):
-        assert self.__uri__ is None
-        self.__uri__ = subject_uri.append(self.method_name)
+    def __repr__(self):
+        return f"MethodSubject {repr(self.subject)}"
+
+
+def is_op_ref(fn):
+    if isinstance(fn, Op):
+        return True
+    elif hasattr(fn, "__form__"):
+        return is_op_ref(form_of(fn))
+    elif isinstance(fn, list) or isinstance(fn, tuple):
+        return any(is_op_ref(item) for item in fn)
+    elif isinstance(fn, dict):
+        return any(is_op_ref(fn[k]) for k in fn)
+    else:
+        return False
+
+
+def is_write_op_ref(fn):
+    if isinstance(fn, Delete) or isinstance(fn, Put):
+        return True
+    elif hasattr(fn, "__form__"):
+        return is_write_op_ref(form_of(fn))
+    elif isinstance(fn, list) or isinstance(fn, tuple):
+        return any(is_write_op_ref(item) for item in fn)
+    elif isinstance(fn, dict):
+        return any(is_write_op_ref(fn[k]) for k in fn)
+    else:
+        return False
+
+
+def log_anonymous(arg):
+    if is_write_op_ref(arg):
+        logging.warning(f"assigning auto-generated name to the result of {arg}")
+    elif is_op_ref(arg):
+        logging.info(f"assigning auto-generated name to the result of {arg}")
 
 
 def reference(cxt, state):
-    name = f"{state.__class__.__name__}_{id(state)}"
-    setattr(cxt, name, state)
+    name = f"{state.__class__.__name__}_{hex_id(state)}"
+    if name not in cxt:
+        logging.debug(f"assigned name {name} to {state} in {cxt}")
+        setattr(cxt, name, state)
+
     return URI(name)
