@@ -1,8 +1,9 @@
 import inspect
+import logging
 
 from tinychain import op, ref
 from tinychain.state import State
-from tinychain.util import form_of, requires, to_json, uri, Context, URI
+from tinychain.util import form_of, to_json, uri, Context, URI
 from tinychain.value import Nil, Value
 
 from . import _get_rtype, resolve_class
@@ -20,11 +21,6 @@ class Op(object):
 
         self.form = form
 
-    def __deps__(self):
-        params = inspect.signature(self.form).parameters
-        provided = set(URI(param) for param in params)
-        return requires(form_of(self)) - provided
-
     def __json__(self):
         return {str(uri(self)): to_json(form_of(self))}
 
@@ -42,8 +38,9 @@ class Get(Op):
         self.rtype = _get_rtype(form, State)
         Op.__init__(self, form)
 
-    def __call__(self, key=None):
-        return ref.Get(self, key)
+    def __dbg__(self):
+        _, cxt = form_of(self)
+        return [cxt]
 
     def __form__(self):
         cxt, args = _maybe_first_arg(self)
@@ -57,14 +54,17 @@ class Get(Op):
             args.append(dtype(URI(key_name)))
 
         cxt._return = self.form(*args)  # populate the Context
+
+        validate(cxt, sig.parameters)
+
         return key_name, cxt
 
     def __ref__(self, name):
         rtype = self.rtype
 
         class GetRef(op.Get):
-            def __call__(self, *args, **kwargs):
-                return rtype(op.Get.__call__(self, *args, **kwargs))
+            def __call__(self, key):
+                return rtype(ref.Get(self, key))
 
         return GetRef(URI(name))
 
@@ -78,6 +78,10 @@ class Put(Op):
     def __call__(self, key=None, value=None):
         return ref.Put(self, key, value)
 
+    def __dbg__(self):
+        _, _, cxt = form_of(self)
+        return [cxt]
+
     def __form__(self):
         cxt, args = _maybe_first_arg(self)
 
@@ -89,9 +93,9 @@ class Put(Op):
             pass
         elif len(sig.parameters) - len(args) == 1:
             param_name = list(sig.parameters.keys())[-1]
-            param = sig.parameters[key_name]
+            param = sig.parameters[param_name]
             dtype = resolve_class(self.form, param.annotation, Value)
-            if param_name in ["key", "value"]:
+            if param_name in set(["key", "value"]):
                 args.append(dtype(URI(param_name)))
             else:
                 raise ValueError(f"{self.dtype()} argument {param_name} is ambiguous--use 'key' or 'value' instead")
@@ -112,12 +116,15 @@ class Put(Op):
                 f"{self.dtype()} accepts a maximum of three arguments: (cxt, key, value) (found {param_names})")
 
         cxt._return = self.form(*args)
+
+        validate(cxt, sig.parameters)
+
         return key_name, value_name, cxt
 
     def __ref__(self, name):
         class PutRef(op.Put):
-            def __call__(self, *args, **kwargs):
-                return Nil(op.Put.__call__(self, *args, **kwargs))
+            def __call__(self, key=None, value=None):
+                return Nil(ref.Put(key, value))
 
         return PutRef(URI(name))
 
@@ -132,14 +139,8 @@ class Post(Op):
         self.rtype = _get_rtype(form, State)
         Op.__init__(self, form)
 
-    def __call__(self, *args, **kwargs):
-        if args and kwargs:
-            raise ValueError("POST Op takes one arg (a Map) or kwargs, but not both")
-
-        if not args:
-            args = kwargs
-
-        return ref.Get(self, args)
+    def __dbg__(self):
+        return [form_of(self)]
 
     def __form__(self):
         cxt, args = _maybe_first_arg(self)
@@ -151,6 +152,9 @@ class Post(Op):
             kwargs[name] = dtype(URI(name))
 
         cxt._return = self.form(*args, **kwargs)
+
+        validate(cxt, sig.parameters)
+
         return cxt
 
     def __ref__(self, name):
@@ -158,7 +162,13 @@ class Post(Op):
 
         class PostRef(op.Post):
             def __call__(self, *args, **kwargs):
-                return rtype(op.Post.__call__(self, *args, **kwargs))
+                if args and kwargs:
+                    raise ValueError("POST Op takes one arg (a Map) or kwargs, but not both")
+
+                if not args:
+                    args = kwargs
+
+                return rtype(ref.Post(self, args))
 
         return PostRef(URI(name))
 
@@ -169,16 +179,17 @@ class Post(Op):
 class Delete(Op):
     __uri__ = uri(op.Delete)
 
-    def __call__(self, key=None):
-        return ref.Get(self, key)
+    def __dbg__(self):
+        _, cxt = form_of(self)
+        return [cxt]
 
     def __form__(self):
         return Get.__form__(self)
 
     def __ref__(self, name):
         class DeleteRef(op.Delete):
-            def __call__(self, *args, **kwargs):
-                return Nil(op.Delete.__call__(self, *args, **kwargs))
+            def __call__(self, key=None):
+                return Nil(ref.Delete(self, key))
 
         return DeleteRef(URI(name))
 
@@ -202,4 +213,38 @@ def _maybe_first_arg(op):
             else:
                 raise ValueError(f"{param} must be a {Context}, not {param.annotation}")
 
+    validate(cxt, sig.parameters)
+
     return cxt, args
+
+
+def validate(cxt, provided):
+    defined = set(provided)
+    for name in cxt.form:
+        def validate_ref(ref):
+            if not hasattr(ref, "__uri__") and not isinstance(ref, URI):
+                return
+
+            ref = uri(ref)
+            if ref.id() is not None and ref.id() not in defined:
+                logging.info(f"{cxt} depends on undefined state {ref.id()}--is it part of a Closure?")
+
+        form = cxt.form[name]
+        while hasattr(form, "__form__"):
+            form = form_of(form)
+
+        if isinstance(form, URI):
+            validate_ref(form)
+        if isinstance(form, ref.Op):
+            validate_ref(form.subject)
+
+            if isinstance(form.args, list) or isinstance(form.args, tuple):
+                for arg in form.args:
+                    validate_ref(arg)
+            elif isinstance(form.args, dict):
+                for arg_name in form.args:
+                    validate_ref(form.args[arg_name])
+            else:
+                validate_ref(form.args)
+
+        defined.add(name)
