@@ -1,4 +1,6 @@
 """TinyChain `State` s, like `Map`, `Tuple`, and `Op`."""
+import inspect
+import logging
 
 from tinychain import ref, reflect
 from tinychain.util import *
@@ -18,15 +20,16 @@ class State(object):
 
         if isinstance(form, URI):
             self.__uri__ = form
-        elif isinstance(form, TypeForm) and form_of(form) is not None:
-            self.__uri__ = uri(form)
         elif reflect.is_ref(form):
             self.__uri__ = uri(form)
 
         reflect.meta.gen_headers(self)
 
-    def __deps__(self):
-        return requires(form_of(self))
+    def __dbg__(self):
+        return [form_of(self)]
+
+    def __eq__(self, _other):
+        raise NotImplementedError("State does not support equality; use a more specific type")
 
     def __json__(self):
         form = form_of(self)
@@ -57,7 +60,7 @@ class State(object):
     def _get(self, name, key=None, rtype=None):
         subject = ref.MethodSubject(self, name)
         op_ref = ref.Get(subject, key)
-        rtype = State if rtype is None else rtype
+        rtype = State if rtype is None or not issubclass(rtype, State) else rtype
         return rtype(op_ref)
 
     def _put(self, name, key=None, value=None):
@@ -67,11 +70,9 @@ class State(object):
         return Nil(ref.Put(subject, key, value))
 
     def _post(self, name, params, rtype):
-        from .value import Nil
-
         subject = ref.MethodSubject(self, name)
         op_ref = ref.Post(subject, params)
-        rtype = Nil if rtype is None else rtype
+        rtype = State if rtype is None or not issubclass(rtype, State) else rtype
         return rtype(op_ref)
 
     def _delete(self, name, key=None):
@@ -83,12 +84,27 @@ class State(object):
     def cast(self, dtype):
         """Attempt to cast this `State` into the given `dtype`."""
 
-        return dtype(ref.Get(uri(dtype), self))
+        # TODO: allow casting to a type known only at run-time
+        if not inspect.isclass(dtype) or not issubclass(dtype, State):
+            raise NotImplementedError("dtype to cast into must be known at compile-time")
+
+        return dtype(ref.Get(dtype, self))
+
+    def copy(self):
+        """Create a new `State` by copying this one."""
+
+        return self._get("copy", rtype=self.__class__)
 
     def dtype(self):
         """Return the native base :class:`Class` of this `State`."""
         return self._get("class", rtype=Class)
 
+    def hash(self):
+        """Return the SHA256 hash of this `State` as an :class:`Id`."""
+
+        from .value import Id
+        return self._get("hash", rtype=Id)
+
     def is_none(self):
         """Return `Bool(true)` if this `State` is :class:`Nil`."""
 
@@ -132,7 +148,17 @@ class Map(State):
 
     def __getitem__(self, key):
         form = form_of(self)
-        rtype = type(form[key]) if isinstance(form, TypeForm) else State
+        if isinstance(form, TypeForm):
+            if key not in form:
+                raise KeyError(f"Map with keys {set(form.keys())} does not contain any entry for {key}")
+
+            rtype = type(form[key])
+            rtype = rtype if issubclass(rtype, State) else State
+            if hasattr(form[key], "__ref__"):
+                return rtype(get_ref(form[key], f"{uri(self)}/{key}"))
+        else:
+            rtype = State
+
         return self._get("", key, rtype)
 
     def __json__(self):
@@ -176,8 +202,11 @@ class Tuple(State):
     __uri__ = uri(State) + "/tuple"
 
     def __init__(self, form):
-        form = TypeForm(None, form) if hasattr(form, "__iter__") else form
-        State.__init__(self, form)
+        if hasattr(form, "__form__") and isinstance(form_of(form), TypeForm):
+            return State.__init__(self, form_of(form))
+        else:
+            form = TypeForm(None, form) if hasattr(form, "__iter__") else form
+            return State.__init__(self, form)
 
     def __add__(self, other):
         return self.extend(other)
@@ -190,7 +219,14 @@ class Tuple(State):
 
     def __getitem__(self, key):
         form = form_of(self)
-        rtype = type(form[key]) if isinstance(form, TypeForm) else State
+        if isinstance(form, TypeForm):
+            rtype = type(form[key])
+            rtype = rtype if issubclass(rtype, State) else State
+            if hasattr(form[key], "__ref__"):
+                return rtype(get_ref(form[key], f"{uri(self)}/{key}"))
+        else:
+            rtype = State
+
         return self._get("", key, rtype)
 
     def __ne__(self, other):
@@ -236,7 +272,7 @@ class Tuple(State):
         """
 
         rtype = type(initial_state) if isinstance(initial_state, State) else State
-        return self._post("fold", Map(item_name=item_name, value=initial_state, op=op), rtype)
+        return self._post("fold", {"item_name": item_name, "value": initial_state, "op": op}, rtype)
 
     def map(self, op):
         """Construct a new `Tuple` by mapping the elements in this `Tuple` with the given `op`."""
@@ -289,11 +325,21 @@ class Stream(State):
     def aggregate(self):
         return self._get("aggregate", rtype=Stream)
 
+    def filter(self, op):
+        """Return a new `Stream` containing only those elements of this `Stream` where the given `op` returns `True`."""
+
+        return self._post("filter", {"op": op}, Stream)
+
     def first(self):
         """Return the first item in this `Stream`, or `Nil` if the `Stream` is empty."""
 
         return self._get("first", rtype=Scalar)
 
+    def flatten(self):
+        """Flatten a `Stream` of `Stream` s into a single `Stream` of their component elements."""
+
+        return self._get("flatten", rtype=Stream)
+
     def for_each(self, op):
         """Run the given `op` for each item in this `Stream`, then return the last result.
 
@@ -301,7 +347,7 @@ class Stream(State):
         """
 
         rtype = op.rtype if hasattr(op, "rtype") else State
-        return self._post("for_each", Map(op=op), rtype)
+        return self._post("for_each", {"op": op}, rtype)
 
     def fold(self, item_name, initial_state, op):
         """Run the given `op` for each item in this `Stream` along with the previous result.
@@ -310,43 +356,57 @@ class Stream(State):
         """
 
         rtype = type(initial_state) if isinstance(initial_state, State) else State
-        return self._post("fold", Map(item_name=item_name, value=initial_state, op=op), rtype)
+        return self._post("fold", {"item_name": item_name, "value": initial_state, "op": op}, rtype)
 
     def map(self, op):
         """Return a new `Stream` whose items are the results of running `op` on each item of this `Stream`."""
 
-        return self._post("map", Map(op=op), Stream)
+        return self._post("map", {"op": op}, Stream)
 
 
 # User-defined object types
 
 class Class(State):
-    """A user-defined TinyChain class."""
+    """A TinyChain class (possibly a user-defined class)."""
 
     __uri__ = uri(State) + "/object/class"
 
+    def __call__(self, *args, **kwargs):
+        if args and kwargs:
+            raise ValueError("Class.__call__ accepts args or kwargs but not both")
+
+        subject = ref.MethodSubject(self)
+        if args:
+            return ref.Get(subject, args)
+        else:
+            return ref.Get(subject, kwargs)
+
 
 class Instance(State):
     """An instance of a user-defined :class:`Class`."""
 
     __uri__ = uri(State) + "/object/instance"
 
+    def copy(self):
+        raise NotImplementedError("abstract method")
+
 
 # Private helper classes
 
 class TypeForm(object):
     def __init__(self, form, type_data):
-        if isinstance(form, URI):
-            self.__uri__ = form
+        while isinstance(type_data, TypeForm):
+            if form is None:
+                form = type_data.__form__
+
+            type_data = type_data.type_data
+
+        if reflect.is_ref(form):
+            self.__uri__ = uri(form)
 
         self.__form__ = form
         self.type_data = type_data
 
-    def __deps__(self):
-        deps = set() if self.__form__ is None else requires(self.__form__)
-        deps.update(requires(self.type_data))
-        return deps
-
     def __getitem__(self, key):
         return self.type_data[key]
 
@@ -365,3 +425,6 @@ class TypeForm(object):
     def __json__(self):
         form = form_of(self)
         return to_json(self.type_data if form is None else form)
+
+    def keys(self):
+        return self.type_data.keys()
