@@ -11,7 +11,7 @@ class Context(object):
     """A transaction context."""
 
     def __init__(self, context=None):
-        object.__setattr__(self, "form", OrderedDict())
+        object.__setattr__(self, "_form", OrderedDict())
 
         if context:
             if isinstance(context, self.__class__):
@@ -33,16 +33,15 @@ class Context(object):
 
         return concat
 
-    def __deps__(self):
-        provided = set(URI(name) for name in self.form.keys())
+    def __contains__(self, item):
+        return self._get_name(item) in self.form
 
-        deps = set()
-        for state in self.form.values():
-            deps.update(requires(state))
-
-        return deps - provided
+    def __dbg__(self):
+        return [self.form[next(reversed(self.form))]] if self.form else []
 
     def __getattr__(self, name):
+        name = self._get_name(name)
+
         if name in self.form:
             value = self.form[name]
             if hasattr(value, "__ref__"):
@@ -53,13 +52,29 @@ class Context(object):
         else:
             raise AttributeError(f"Context has no such value: {name}")
 
+    def __getitem__(self, selector):
+        if isinstance(selector, slice):
+            cxt = Context()
+            keys = list(self.form.keys())
+            for name in keys[selector]:
+                setattr(cxt, name, self.form[name])
+
+            return cxt
+        else:
+            return getattr(self, selector)
+
     def __json__(self):
         return to_json(list(self.form.items()))
 
+    def __len__(self):
+        return len(self.form)
+
     def __setattr__(self, name, state):
         if state is self:
             raise ValueError(f"cannot assign transaction Context to itself")
 
+        name = self._get_name(name)
+
         deanonymize(state, self)
 
         if name in self.form:
@@ -69,17 +84,30 @@ class Context(object):
 
     def __repr__(self):
         data = list(self.form.keys())
-        return f"execution context with data {data}"
+        return f"Op context with data {data}"
 
-    def is_defined(self, name):
-        if isinstance(name, URI):
-            name = name.id()
+    def _get_name(self, item):
+        if hasattr(item, "__uri__"):
+            if uri(item).id() != uri(item):
+                raise ValueError(f"invalid name: {item}")
+            else:
+                return uri(item).id()
+        else:
+            return str(item)
 
-        return name in self.form
+    @property
+    def form(self):
+        return self._form
+
+
+def debug(state):
+    """Return the name and immediate dependencies of the given `state`."""
+
+    return [dep for dep in state.__dbg__() if dep is not None] if hasattr(state, "__dbg__") else []
 
 
 def form_of(state):
-    """Return the form of the given state."""
+    """Return the form of the given `state`."""
 
     if hasattr(state, "__form__"):
         if callable(state.__form__) and not inspect.isclass(state.__form__):
@@ -140,7 +168,8 @@ class URI(object):
     """
 
     def __init__(self, root, path=[]):
-        assert root is not None
+        if not root:
+            raise ValueError(f"invalid URI root: {root}")
 
         if root.startswith("$$"):
             raise ValueError(f"invalid reference: {root}")
@@ -157,14 +186,6 @@ class URI(object):
     def __radd__(self, other):
         return URI(other) + str(self)
 
-    def __deps__(self):
-        if "://" in self._root or self.startswith('/'):
-            return set()
-        elif '/' in self._root:
-            return set([URI(self._root[:self._root.index('/')])])
-        else:
-            return set([URI(self._root)])
-
     def __eq__(self, other):
         return str(self) == str(other)
 
@@ -206,6 +227,17 @@ class URI(object):
 
         return URI(str(self), [name])
 
+    def id(self):
+        """Return the ID segment of this `URI`, if present."""
+
+        this = str(self)
+        if this.startswith('$'):
+            if '/' in this:
+                end = this.index("/")
+                return this[1:end]
+            else:
+                return this[1:]
+
     def is_id(self):
         """Return `True` if this URI is a simple ID, like `$foo`."""
 
@@ -266,20 +298,6 @@ class URI(object):
             if i > 0:
                 return self._root[:i]
 
-    def subject(self):
-        """Return only the ID portion of this `URI`, or `None` in the case of a link."""
-
-        if "://" in self._root:
-            return None
-        elif self._root.startswith('/'):
-            return None
-
-        uri = str(self)
-        if '/' in uri:
-            return URI(uri[:uri.index('/')])
-        else:
-            return URI(uri)
-
     def startswith(self, prefix):
         return str(self).startswith(str(prefix))
 
@@ -288,28 +306,6 @@ def print_json(obj):
     print(json.dumps(to_json(obj), indent=4))
 
 
-def requires(subject):
-    """Return a set of the IDs of the dependencies required to resolve the given state."""
-
-    if inspect.isclass(subject):
-        return set()
-
-    if hasattr(subject, "__deps__"):
-        return subject.__deps__()
-
-    deps = set()
-
-    if isinstance(subject, list) or isinstance(subject, tuple):
-        for item in subject:
-            deps.update(requires(item))
-
-    elif isinstance(subject, dict):
-        for item in subject.values():
-            deps.update(requires(item))
-
-    return deps
-
-
 def uri(subject):
     """Return the `URI` of the given state."""
 
@@ -337,11 +333,15 @@ def use(cls):
 def to_json(obj):
     """Return a JSON-encodable representation of the given state or reference."""
 
+    err_native_method = "Python method {} is not JSON serializable; consider using a decorator like @get_method"
+    err_native_function = "Python function {} is not JSON serializable; consider using a decorator like @get_op"
+
     if inspect.ismethod(obj):
         if not hasattr(obj, "__json__"):
-            raise ValueError(
-                f"Python method {obj} is not JSON serializable; "
-                + "try using a decorator like @get_method")
+            raise ValueError(err_native_method.format(obj))
+    elif inspect.isfunction(obj):
+        if not hasattr(obj, "__json__"):
+            raise ValueError(err_native_function.format(obj))
 
     if inspect.isclass(obj):
         if hasattr(type(obj), "__json__"):
