@@ -14,10 +14,14 @@ class Value(Scalar):
 
     @classmethod
     def max(cls, l, r):
+        """Return `l`, or `r`, whichever is greater."""
+
         return cls(If(l >= r, l, r))
 
     @classmethod
     def min(cls, l, r):
+        """Return `l`, or `r`, whichever is lesser."""
+
         return cls(If(l <= r, l, r))
 
     def __eq__(self, other):
@@ -92,6 +96,12 @@ class Bytes(Value):
     __uri__ = uri(Value) + "/bytes"
 
 
+class EmailAddress(Value):
+    """An email address"""
+
+    __uri__ = uri(Value) + "/email"
+
+
 class Id(Value):
     """An identifier"""
 
@@ -156,6 +166,9 @@ class Number(Value):
     def __div__(self, other):
         return self.div(other)
 
+    def __mod__(self, other):
+        return self.modulo(other)
+
     def __mul__(self, other):
         return self.mul(other)
 
@@ -238,6 +251,20 @@ class Number(Value):
 
         return self._get("div", other, self.__class__)
 
+    def log(self, base=None):
+        """
+        Return the logarithm of this `Number`.
+
+        If no `base` is specified, this will be the natural logarithm (base e).
+        """
+
+        return self._post("log", {"r": base}, F64)
+
+    def modulo(self, other):
+        """Return the remainder of `self` divided by `other`."""
+
+        return self._get("mod", other, self.__class__)
+
     def mul(self, other):
         """Return the product of `self` and `other`."""
 
@@ -248,6 +275,15 @@ class Number(Value):
 
         return self._get("pow", other, self.__class__)
 
+    def round(self, digits=None):
+        """Round this `Number` to the nearest integer, or `digits` decimal places (if `digits` is provided)."""
+
+        if digits is None:
+            return self._get("round", rtype=self.__class__)
+        else:
+            places = Int(10)**digits
+            return (self * places).round() / places
+
     def sin(self):
         """Return the sine of this `Number`."""
 
@@ -361,10 +397,14 @@ class Int(Number):
 
     @classmethod
     def max_value(cls):
+        """Return the maximum allowed value of this `Int` type."""
+
         return cls(2**(cls.size() - 1) - 1)
 
     @classmethod
     def min_value(cls):
+        """Return the minimum allowed value of this `Int` type."""
+
         return cls(-2**(cls.size() - 1) + 1)
 
     @staticmethod
@@ -417,10 +457,14 @@ class UInt(Number):
 
     @classmethod
     def max_value(cls):
+        """Return the maximum allowed value of this `Int` type."""
+
         return cls(2**cls.size() - 1)
 
     @classmethod
     def min_value(cls):
+        """Return the minimum allowed value of this `Int` type."""
+
         return cls(0)
 
     @staticmethod
